from web3.providers import HTTPProvider
from conflux import Conflux
from cfx_account.account import Account
from hexbytes import (
    HexBytes,
)

def print_fields(item):
    for key in item:
        value = item[key]
        if isinstance(value, HexBytes):
            value = value.hex()
        print(key, ':', value)


conflux_testnet_rpc = 'https://testnet-rpc.conflux-chain.org.cn/v2'
provider = HTTPProvider(conflux_testnet_rpc)
conflux = Conflux(provider)
account = "cfxtest:aak7fsws4u4yf38fk870218p1h3gxut3ku00u1k1da"
tx_hash = "0xbe2a7cafb652f4c95a8a6fb3f2dc72ec9ff914fe27bcd1280b6be905481e5473"
block_hash = "0x573f11c723e98cfbdbfcc3d32b1d6b55d5e2d94082e13ede1866dc47ddda2359"
epoch_number = 19978013
private_key = '0xcc7939276283a32f60d2fad7d16cac972300308fe99ec98d0e63765d02e24863'
address = 'cfxtest:aar3uh6bm4hr5bb73rrya99u5y1cm2pgeja196rfeb'
key_account = Account.from_key(private_key)

txinfo = {
    "from": address,
    "to": "cfxtest:aak7fsws4u4yf38fk870218p1h3gxut3ku00u1k1da",
    "value": 10
}

block_fields = [
    "adaptive",
    "blame",
    "deferredLogsBloomHash",
    "deferredReceiptsRoot",
    "deferredStateRoot",
    "difficulty",
    "epochNumber",
    "gasLimit",
    "gasUsed",
    "hash",
    "height",
    "miner",
    "nonce",
    "parentHash",
    "powQuality",
    "refereeHashes",
    "size",
    "timestamp",
    "transactions",
    "transactionsRoot"
]

tx_fields = [
    'blockHash',
    'chainId',
    'contractCreated',
    'data',
    'epochHeight',
    'from',
    'gas',
    'gasPrice',
    'hash',
    'nonce',
    'r', 's', 'v',
    'status',
    'storageLimit',
    'to',
    'transactionIndex',
    'value'
]

receipt_fields = [
    'transactionHash',
    'index',
    'blockHash',
    'epochNumber',
    'from', 'to',
    'gasUsed', 'gasFee',
    'gasCoveredBySponsor',
    'storageCollateralized',
    'storageCoveredBySponsor',
    'storageReleased',
    'contractCreated',
    'stateRoot',
    'outcomeStatus',
    'logsBloom',
    'logs'
]

status_fields = [
    'networkId', 'chainId',
    'blockNumber', 'epochNumber',
    'bestHash',
    'pendingTxNumber',
    'latestState', 'latestCheckpoint', 'latestConfirmed'
]

account_fields = [
    'balance', 'stakingBalance',
    'nonce',
    'admin', 'codeHash',
    'collateralForStorage',
    'accumulatedInterestReturn'
]

estimate_fields = [
    'gasLimit',
    'gasUsed',
    'storageCollateralized'
]

def check_fields(to_check, fields):
    for f in fields:
        assert f in to_check

def test_getStatus():
    status = conflux.cfx.getStatus()
    check_fields(status, status_fields)
    assert status['networkId'] == 1
    assert status['chainId'] == 1
    assert type(status['blockNumber']) == int
    assert type(status['epochNumber']) == int
    assert type(status['pendingTxNumber']) == int
    assert type(status['latestCheckpoint']) == int
    assert type(status['latestConfirmed']) == int
    assert type(status['latestState']) == int
    assert type(status['bestHash']) == str

def test_gasPrice():
    price = conflux.cfx.gasPrice
    assert type(price) == int

def test_getNextNonce():
    nonce = conflux.cfx.getNextNonce(account)
    assert type(nonce) == int

def test_getBalance():
    balance = conflux.cfx.getBalance(account)
    assert type(balance) == int

def test_getStakingBalance():
    balance = conflux.cfx.getStakingBalance(account)
    assert type(balance) == int

def test_epochNumber():
    epoch_number = conflux.cfx.epochNumber()
    assert type(epoch_number) == int

def test_account():
    account_info = conflux.cfx.getAccount(account)
    check_fields(account_info, account_fields)
    assert type(account_info['balance']) == int
    assert type(account_info['nonce']) == int
    assert type(account_info['stakingBalance']) == int
    assert type(account_info['collateralForStorage']) == int
    assert type(account_info['accumulatedInterestReturn']) == int
    assert type(account_info['codeHash']) == str
    assert type(account_info['admin']) == str

def test_getBlockByHash():
    block = conflux.cfx.getBlockByHash(block_hash)
    check_fields(block, block_fields)
    assert type(block['adaptive']) == bool

def test_getBlockByEpochNumber():
    block = conflux.cfx.getBlockByEpochNumber(epoch_number)
    check_fields(block, block_fields)

def test_getTransactionByHash():
    tx = conflux.cfx.getTransactionByHash(tx_hash)
    check_fields(tx, tx_fields)

def test_getTransactionReceipt():
    receipt = conflux.cfx.getTransactionReceipt(tx_hash)
    check_fields(receipt, receipt_fields)

def test_getLogs():
    pass

def test_call():
    pass

def test_estimate():
    estimate = conflux.cfx.estimateGasAndCollateral(txinfo)
    check_fields(estimate, estimate_fields)

def test_sendRawTransaction():
    conflux.cfx.populate_transaction(txinfo)
    signed_tx = Account.sign_transaction(txinfo, private_key)
    tx_hash = conflux.cfx.sendRawTransaction(signed_tx.rawTransaction.hex())
    print(tx_hash)



from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    NewType,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
)
from eth_typing import (
    BlockNumber,
    Hash32,
    HexStr,
)
from hexbytes import (
    HexBytes,
)
from web3._utils.compat import (
    Literal,
    TypedDict,
)
from web3.types import (
    Nonce,
    Timestamp,
    _Hash32,
)

# if TYPE_CHECKING:
#     from web3 import Web3  # noqa: F401

EpochTags = Literal["earliest", "latest_checkpoint", "latest_confirmed", "latest_state", "latest_mined"]
EpochIdentifier = Union[EpochTags, int, HexStr]
BlockIdentifier = Union[EpochIdentifier, Hash32, HexStr, HexBytes]
LatestEpochParam = Literal["latest_state"]

Drip = NewType('Drip', int)
EpochNumber = NewType("EpochNumber", int)

class FilterParams(TypedDict, total=False):
    address: Union[str, List[str]]
    blockHashes: List[HexBytes]
    fromEpoch: EpochIdentifier
    toEpoch: EpochIdentifier
    topics: Sequence[Optional[Union[_Hash32, Sequence[_Hash32]]]]
    limit: int

class LogReceipt(TypedDict):
    address: str
    data: HexStr
    topics: Sequence[HexBytes]

class ChainStatus:
    bestHash: str
    blockNumber: int
    chainId: int
    networkId: int
    epochNumber: int
    pendingTxNumber: int
    latestCheckpoint: int
    latestConfirmed: int
    latestState: int

class Account:
    balance: int
    nonce: int
    codeHash: str
    admin: str
    stakingBalance: int
    collateralForStorage: int
    accumulatedInterestReturn: int

class Log(TypedDict):
    address: str
    topics: Sequence[HexBytes]
    data: HexStr
    blockHash: HexBytes
    epochNumber: BlockNumber
    transactionHash: HexBytes
    transactionIndex: int
    logIndex: int
    transactionLogIndex: int

TxData = TypedDict("TxData", {
    "blockHash": HexBytes,
    "contractCreated": str,
    "epochHeight": BlockNumber,
    "chainId": int,
    "data": Union[bytes, HexStr],
    "from": str,
    "gas": Drip,
    "gasPrice": Drip,
    "hash": HexBytes,
    "nonce": Nonce,
    "r": HexBytes,
    "s": HexBytes,
    "status": int,
    "storageLimit": int,
    "to": str,
    "transactionIndex": int,
    "v": int,
    "value": Drip,
}, total=False)

TxParams = TypedDict("TxParams", {
    "chainId": int,
    "data": Union[bytes, HexStr],
    "from": str,
    "gas": Drip,
    "gasPrice": Drip,
    "nonce": Nonce,
    "to": str,
    "value": Drip,
    "epochHeight": int,
    "storageLimit": int,
}, total=False)

class StorageRelease:
    address: str
    collaterals: int

TxReceipt = TypedDict("TxReceipt", {
    "blockHash": HexBytes,
    "epochNumber": BlockNumber,
    "gasUsed": Drip,
    "gasFee": Drip,
    "gasCoveredBySponsor": bool,
    "storageCoveredBySponsor": bool,
    "storageCollateralized": int,
    "storageReleased": List[StorageRelease],
    "contractCreated": str,
    "stateRoot": HexBytes,
    "from": str,
    "logs": List[LogReceipt],
    "logsBloom": HexBytes,
    "outcomeStatus": int,
    "to": str,
    "transactionHash": HexBytes,
    "index": int,
})

class BlockData(TypedDict, total=False):
    adaptive: bool
    blame: int
    deferredLogsBloomHash: HexBytes
    deferredReceiptsRoot: HexBytes
    deferredStateRoot: HexBytes
    difficulty: int
    gasLimit: Drip
    powQuality: HexBytes
    refereeHashes: List[HexBytes]
    gasUsed: Drip
    hash: HexBytes
    miner: str
    nonce: HexBytes
    epochNumber: BlockNumber
    height: BlockNumber
    parentHash: HexBytes
    size: int
    timestamp: Timestamp
    transactions: Union[Sequence[HexBytes], Sequence[TxData]]
    transactionsRoot: HexBytes

class EstimateResult(TypedDict):
    gasUsed: int
    storageCollateralized: int

class SponsorInfo(TypedDict):
    sponsorBalanceForCollateral: int
    sponsorBalanceForGas: int
    sponsorGasBound: int
    sponsorForCollateral: str
    sponsorForGas: str

class RewardInfo:
    author: str
    baseReward: int
    blockHash: str
    totalReward: int
    txFee: int


from conflux._utils.rpc_abi import (
    RPC,
)
from web3.manager import (
    RequestManager as DefaultRequestManager,
)
from web3.providers import (
    BaseProvider,
)
from web3.providers.ipc import (
    IPCProvider,
)
from web3.providers.rpc import (
    HTTPProvider,
)
from web3.providers.websocket import (
    WebsocketProvider,
)
from typing import (
    Any,
    cast,
    Dict,
    List,
    Optional,
    Sequence,
    TYPE_CHECKING
)
from web3._utils.module import (
    attach_modules,
)
from eth_abi.codec import (
    ABICodec,
)
from web3._utils.abi import (
    build_default_registry,
    build_strict_registry,
    map_abi_data,
)
from conflux.cfx import Cfx
from eth_utils import (
    to_wei,
    from_wei,
)
from eth_utils.address import (
    to_checksum_address
)
from eth_utils.conversions import (
    to_bytes
)
from web3._utils.contracts import (
    find_matching_fn_abi
)
from web3._utils.abi import (
    get_abi_output_types
)
from cfx_address import (
    Address
)
from web3 import (
    Web3,
)

def get_default_modules() -> Dict[str, Sequence[Any]]:
    return {
        "cfx": (Cfx,),
    }

class Conflux:
    # Providers
    HTTPProvider = HTTPProvider
    IPCProvider = IPCProvider
    WebsocketProvider = WebsocketProvider

    # Managers
    RequestManager = DefaultRequestManager

    # Currency Utility
    toDrip = staticmethod(to_wei)
    fromDrip = staticmethod(from_wei)

    cfx: Cfx

    def __init__(
        self,
        provider: Optional[BaseProvider] = None,
        middlewares: Optional[Sequence[Any]] = None,
        modules: Optional[Dict[str, Sequence[Any]]] = None,
    ) -> None:
        self.manager = self.RequestManager(self, provider, middlewares)

        self.codec = ABICodec(build_default_registry())

        if modules is None:
            modules = get_default_modules()

        attach_modules(self, modules)

        self._w3 = Web3(Web3.EthereumTesterProvider())

    @property
    def clientVersion(self) -> str:
        return self.manager.request_blocking(RPC.cfx_clientVersion, [])

    def contract(self, address, abi):
        hex_address = address
        if Address.has_network_prefix(address):
            hex_address = Address(address).eth_checksum_address

        return self._w3.eth.contract(address=hex_address, abi=abi)

    def call_contract_method(self, address, abi, method_name, *kargs):
        contract = self.contract(address, abi)
        tx = contract.functions[method_name](*kargs).buildTransaction({
            "gas": 21000,
            "gasPrice": 1,
        })
        call_result = self.cfx.call({
            "to": address,
            "data": tx['data']
        })
        fn_abi = find_matching_fn_abi(contract.abi, self._w3.codec, method_name, kargs)
        output_types = get_abi_output_types(fn_abi)
        decoded_result = self._w3.codec.decode_abi(output_types, to_bytes(hexstr=call_result))
        if len(output_types) == 1:
            return decoded_result[0]
        else:
            return decoded_result



from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Collection,
    Dict,
    Iterable,
    NoReturn,
    Tuple,
    Union,
)
from web3.types import (
    BlockIdentifier,
    RPCEndpoint,
    RPCResponse,
    TReturn,
    _Hash32,
)
from eth_utils.curried import (
    apply_formatter_at_index,
    apply_formatter_if,
    apply_formatter_to_array,
    apply_formatters_to_dict,
    apply_formatters_to_sequence,
    apply_one_of_formatters,
    is_0x_prefixed,
    is_address,
    is_bytes,
    is_dict,
    is_integer,
    is_null,
    is_string,
    remove_0x_prefix,
    text_if_str,
    to_checksum_address,
    to_list,
    to_tuple,
)
from web3._utils.method_formatters import (
    combine_formatters,
    compose,
    not_attrdict,
    apply_module_to_formatters,
    to_integer_if_hex,
    to_hex_if_integer,
    is_not_false,
    is_false,
    is_null,
    is_not_null,
    to_hexbytes,
    apply_list_to_array_formatter,
    is_array_of_dicts,
    is_array_of_strings,
    to_ascii_if_bytes,
)
from web3.datastructures import (
    AttributeDict,
)
from hexbytes import (
    HexBytes,
)
from conflux._utils.rpc_abi import (
    RPC,
    RPC_ABIS,
)
from web3._utils.rpc_abi import (
    abi_request_formatters,
)
from web3._utils.normalizers import (
    abi_address_to_hex,
    abi_bytes_to_hex,
    abi_int_to_hex,
    abi_string_to_hex,
)

from web3._utils.formatters import (
    integer_to_hex
)

if TYPE_CHECKING:
    from web3 import Web3  # noqa: F401
    from web3.module import (  # noqa: F401
        Module,
        ModuleV2,
    )

FILTER_RESULT_FORMATTERS: Dict[RPCEndpoint, Callable[..., Any]] = {
    # RPC.eth_newPendingTransactionFilter: filter_wrapper,
    # RPC.eth_newBlockFilter: filter_wrapper,
    # RPC.eth_newFilter: filter_wrapper,
}

FILTER_PARAM_NORMALIZERS = apply_formatters_to_dict({
    'address': apply_formatter_if(is_string, lambda x: [x])
})

STANDARD_NORMALIZERS = [
    abi_bytes_to_hex,
    abi_int_to_hex,
    abi_string_to_hex,
    abi_address_to_hex,
]

STATUS_FORMATTERS = {
    'chainId': to_integer_if_hex,
    'networkId': to_integer_if_hex,
    'blockNumber': to_integer_if_hex,
    'epochNumber': to_integer_if_hex,
    'pendingTxNumber': to_integer_if_hex,
    'latestCheckpoint': to_integer_if_hex,
    'latestConfirmed': to_integer_if_hex,
    'latestState': to_integer_if_hex,
}

status_formatter = apply_formatters_to_dict(STATUS_FORMATTERS)

ACCOUNT_FORMATTERS = {
    'balance': to_integer_if_hex,
    'nonce': to_integer_if_hex,
    'stakingBalance': to_integer_if_hex,
    'collateralForStorage': to_integer_if_hex,
    'accumulatedInterestReturn': to_integer_if_hex,
}

account_formatter = apply_formatters_to_dict(ACCOUNT_FORMATTERS)

TRANSACTION_FORMATTERS = {
    'blockHash': apply_formatter_if(is_not_null, to_hexbytes(32)),
    'chainId': apply_formatter_if(is_not_null, to_integer_if_hex),
    'epochHeight': apply_formatter_if(is_not_null, to_integer_if_hex),
    'transactionIndex': apply_formatter_if(is_not_null, to_integer_if_hex),
    'nonce': to_integer_if_hex,
    'gas': to_integer_if_hex,
    'gasPrice': to_integer_if_hex,
    'storageLimit': to_integer_if_hex,
    'status': apply_formatter_if(is_not_null, to_integer_if_hex),
    'value': to_integer_if_hex,
    'publicKey': apply_formatter_if(is_not_null, to_hexbytes(64)),
    'raw': HexBytes,
    # 'from': to_checksum_address,
    # 'to': apply_formatter_if(is_address, to_checksum_address),
    'hash': to_hexbytes(32),
    'r': apply_formatter_if(is_not_null, to_hexbytes(32, variable_length=True)),
    's': apply_formatter_if(is_not_null, to_hexbytes(32, variable_length=True)),
    'v': apply_formatter_if(is_not_null, to_integer_if_hex),
}

transaction_formatter = apply_formatters_to_dict(TRANSACTION_FORMATTERS)

BLOCK_FORMATTERS = {
    'blame': to_integer_if_hex,
    'deferredReceiptsRoot': to_hexbytes(32),
    'deferredLogsBloomHash': to_hexbytes(32),
    'deferredStateRoot': to_hexbytes(32),
    'gasLimit': to_integer_if_hex,
    'powQuality': to_integer_if_hex,
    'height': to_integer_if_hex,
    'gasUsed': to_integer_if_hex,
    'size': to_integer_if_hex,
    'timestamp': to_integer_if_hex,
    'hash': apply_formatter_if(is_not_null, to_hexbytes(32)),
    'nonce': apply_formatter_if(is_not_null, to_hexbytes(8, variable_length=True)),
    # 'miner': apply_formatter_if(is_not_null, to_checksum_address),
    # 'nonce': to_integer_if_hex,
    # 'refereeHashes':  array hash
    'epochNumber': apply_formatter_if(is_not_null, to_integer_if_hex),
    'parentHash': apply_formatter_if(is_not_null, to_hexbytes(32)),
    'difficulty': to_integer_if_hex,
    'transactions': apply_one_of_formatters((
        (is_array_of_dicts, apply_list_to_array_formatter(transaction_formatter)),
        (is_array_of_strings, apply_list_to_array_formatter(to_hexbytes(32))),
    )),
    'transactionsRoot': to_hexbytes(32),
}

block_formatter = apply_formatters_to_dict(BLOCK_FORMATTERS)

LOG_ENTRY_FORMATTERS = {
    # 'address': str,
    'topics': apply_list_to_array_formatter(to_hexbytes(32)),
    'data': to_ascii_if_bytes,
}

log_entry_formatter = apply_formatters_to_dict(LOG_ENTRY_FORMATTERS)

LOG_FORMATTERS = {
    'blockHash': apply_formatter_if(is_not_null, to_hexbytes(32)),
    'epochNumber': apply_formatter_if(is_not_null, to_integer_if_hex),
    'transactionIndex': apply_formatter_if(is_not_null, to_integer_if_hex),
    'transactionLogIndex': apply_formatter_if(is_not_null, to_integer_if_hex),
    'transactionHash': apply_formatter_if(is_not_null, to_hexbytes(32)),
    'logIndex': to_integer_if_hex,
    'address': str,
    'topics': apply_list_to_array_formatter(to_hexbytes(32)),
    'data': to_ascii_if_bytes,
}

log_formatter = apply_formatters_to_dict(LOG_FORMATTERS)

logs_result_formatter = apply_list_to_array_formatter(log_entry_formatter)

STORAGE_RELEASE_FORMATTERS = {
    # 'address'
    'collaterals': to_integer_if_hex
}

storage_release_formatt = apply_formatters_to_dict(STORAGE_RELEASE_FORMATTERS)

RECEIPT_FORMATTERS = {
    'blockHash': apply_formatter_if(is_not_null, to_hexbytes(32)),
    'epochNumber': apply_formatter_if(is_not_null, to_integer_if_hex),
    'index': apply_formatter_if(is_not_null, to_integer_if_hex),
    'transactionHash': to_hexbytes(32),
    'status': to_integer_if_hex,
    'gasUsed': to_integer_if_hex,
    'outcomeStatus': to_integer_if_hex,
    'storageCollateralized': to_integer_if_hex,
    'gasFee': to_integer_if_hex,
    'contractAddress': apply_formatter_if(is_not_null, to_checksum_address),
    'logs': apply_list_to_array_formatter(log_entry_formatter),
    'logsBloom': to_hexbytes(256),
    # 'from': apply_formatter_if(is_not_null, to_checksum_address),
    # 'to': apply_formatter_if(is_address, to_checksum_address),
}

receipt_formatter = apply_formatters_to_dict(RECEIPT_FORMATTERS)

ABI_REQUEST_FORMATTERS = abi_request_formatters(STANDARD_NORMALIZERS, RPC_ABIS)

METHOD_NORMALIZERS: Dict[RPCEndpoint, Callable[..., Any]] = {
    RPC.cfx_getLogs: apply_formatter_at_index(FILTER_PARAM_NORMALIZERS, 0),
}

FILTER_PARAMS_FORMATTERS = {
    'fromEpoch': apply_formatter_if(is_integer, integer_to_hex),
    'toEpoch': apply_formatter_if(is_integer, integer_to_hex),
    'limit': apply_formatter_if(is_integer, integer_to_hex),
}

filter_params_formatter = apply_formatters_to_dict(FILTER_PARAMS_FORMATTERS)

ESTIMATE_FORMATTERS = {
    'gasUsed': to_integer_if_hex,
    'storageCollateralized': to_integer_if_hex,
    'gasLimit': to_integer_if_hex,
}

estimate_formatter = apply_formatters_to_dict(ESTIMATE_FORMATTERS)

SPONSOR_INFO_FORMATTERS = {
    'sponsorBalanceForCollateral': to_integer_if_hex,
    'sponsorBalanceForGas': to_integer_if_hex,
    'sponsorGasBound': to_integer_if_hex,
}

sponsor_formatter = apply_formatters_to_dict(SPONSOR_INFO_FORMATTERS)

REWARD_INFO_FORMATTERS = {
    'baseReward': to_integer_if_hex,
    'totalReward': to_integer_if_hex,
    'txFee': to_integer_if_hex,
}

reward_info_formatter = apply_formatters_to_dict(REWARD_INFO_FORMATTERS)

PYTHONIC_REQUEST_FORMATTERS: Dict[RPCEndpoint, Callable[..., Any]] = {
    RPC.cfx_getBalance: apply_formatter_at_index(to_hex_if_integer, 1),
    RPC.cfx_getStakingBalance: apply_formatter_at_index(to_hex_if_integer, 1),
    RPC.cfx_getCollateralForStorage: apply_formatter_at_index(to_hex_if_integer, 1),
    RPC.cfx_getNextNonce: apply_formatter_at_index(to_hex_if_integer, 1),
    RPC.cfx_getAccount: apply_formatter_at_index(to_hex_if_integer, 1),
    RPC.cfx_getAdmin: apply_formatter_at_index(to_hex_if_integer, 1),
    RPC.cfx_getCode: apply_formatter_at_index(to_hex_if_integer, 1),
    RPC.cfx_getSponsorInfo: apply_formatter_at_index(to_hex_if_integer, 1),
    RPC.cfx_getBlockByEpochNumber: apply_formatter_at_index(to_hex_if_integer, 0),
    RPC.cfx_getBlocksByEpoch: apply_formatter_at_index(to_hex_if_integer, 0),
    RPC.cfx_getBlockRewardInfo: apply_formatter_at_index(to_hex_if_integer, 0),
    RPC.cfx_getLogs: apply_formatter_at_index(filter_params_formatter, 0),
}

PYTHONIC_RESULT_FORMATTERS: Dict[RPCEndpoint, Callable[..., Any]] = {
    RPC.cfx_gasPrice: to_integer_if_hex,
    RPC.cfx_epochNumber: to_integer_if_hex,
    RPC.cfx_getBalance: to_integer_if_hex,
    RPC.cfx_getStakingBalance: to_integer_if_hex,
    RPC.cfx_getCollateralForStorage: to_integer_if_hex,
    RPC.cfx_getConfirmationRiskByHash: apply_formatter_if(is_not_null, to_integer_if_hex),
    RPC.cfx_getNextNonce: to_integer_if_hex,
    RPC.cfx_getStatus: apply_formatter_if(is_not_false, status_formatter),
    RPC.cfx_getSponsorInfo: apply_formatter_if(is_not_false, sponsor_formatter),
    RPC.cfx_getBlockRewardInfo: apply_formatter_if(is_not_false, reward_info_formatter),
    RPC.cfx_getAccount: apply_formatter_if(is_not_false, account_formatter),
    RPC.cfx_getBlockByHash: apply_formatter_if(is_not_null, block_formatter),
    RPC.cfx_getBlockByEpochNumber: apply_formatter_if(is_not_null, block_formatter),
    RPC.cfx_getTransactionByHash: apply_formatter_if(is_not_null, transaction_formatter),
    RPC.cfx_getTransactionReceipt: apply_formatter_if(is_not_null, receipt_formatter),
    RPC.cfx_getLogs: logs_result_formatter,
    RPC.cfx_estimateGasAndCollateral: apply_formatter_if(is_not_null, estimate_formatter),
    RPC.cfx_sendRawTransaction: to_hexbytes(32),
    RPC.cfx_sendTransaction: to_hexbytes(32),
}

def get_request_formatters(
    method_name: Union[RPCEndpoint, Callable[..., RPCEndpoint]]
) -> Dict[str, Callable[..., Any]]:
    request_formatter_maps = (
        ABI_REQUEST_FORMATTERS,
        # METHOD_NORMALIZERS needs to be after ABI_REQUEST_FORMATTERS
        # so that eth_getLogs's apply_formatter_at_index formatter
        # is applied to the whole address
        # rather than on the first byte of the address
        METHOD_NORMALIZERS,
        PYTHONIC_REQUEST_FORMATTERS,
    )
    formatters = combine_formatters(request_formatter_maps, method_name)
    return compose(*formatters)


def get_result_formatters(
    method_name: Union[RPCEndpoint, Callable[..., RPCEndpoint]],
    module: Union["Module", "ModuleV2"],
) -> Dict[str, Callable[..., Any]]:
    formatters = combine_formatters(
        (PYTHONIC_RESULT_FORMATTERS,),
        method_name
    )
    formatters_requiring_module = combine_formatters(
        (FILTER_RESULT_FORMATTERS,),
        method_name
    )

    partial_formatters = apply_module_to_formatters(
        formatters_requiring_module,
        module,
        method_name
    )
    attrdict_formatter = apply_formatter_if(is_dict and not_attrdict, AttributeDict.recursive)
    return compose(*partial_formatters, attrdict_formatter, *formatters)


from conflux import (
    Conflux,
    HTTPProvider,
)
import json
from cfx_address import Address
import os
current_path = os.path.dirname(os.path.abspath(__file__))

provider = HTTPProvider('https://test.confluxrpc.com')
c = Conflux(provider)

with open(current_path + os.path.sep + 'crc20_abi.json') as json_file:
    abi_json = json.load(json_file)

token_address = 'cfxtest:acattk1yzt7b8z01m8urnmytst1z0wv0tuvczsyr32'
user_address = 'cfxtest:aak2rra2njvd77ezwjvx04kkds9fzagfe6d5r8e957'

def test_contract_call():
    balance = c.call_contract_method(token_address, abi_json["abi"], "balanceOf", Address(user_address).eth_checksum_address)
    print(balance)

def test_contract_update():
    contract = c.contract(token_address, abi_json["abi"])
    data = contract.encodeABI(fn_name="transfer", args=["0x13d2bA4eD43542e7c54fbB6c5fCCb9f269C1f94C", 100])
    print(data)


from web3.module import (
    Module,
    ModuleV2,
)
from web3.method import (
    Method,
    default_root_munger,
)
from typing import (
    Any,
    Callable,
    List,
    NoReturn,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    cast,
    overload,
)
from conflux.types import (
    Drip,
    EpochNumber,
    EpochIdentifier,
    BlockIdentifier,
    BlockData,
    TxData,
    _Hash32,
    TxReceipt,
    TxParams,
    LogReceipt,
    FilterParams,
    Log,
    ChainStatus,
    Account,
    EstimateResult,
    SponsorInfo,
    RewardInfo
)
from conflux._utils.rpc_abi import RPC
from web3._utils.compat import (
    Literal,
)
from hexbytes import (
    HexBytes,
)
from conflux._utils.method_formatters import (
    get_result_formatters,
    get_request_formatters
)
from eth_typing import (
    # Address,
    # BlockNumber,
    # ChecksumAddress,
    HexStr,
)

from conflux.consts import (
    LATEST_STATE,
    LATEST_MINED
)

class Cfx(ModuleV2, Module):
    defaultEpoch: EpochIdentifier = LATEST_STATE

    gas_price: Method[Callable[[], Drip]] = Method(
        RPC.cfx_gasPrice,
        mungers=None,
        result_formatters = get_result_formatters
    )

    @property
    def gasPrice(self) -> Drip:
        return self.gas_price()

    get_accounts: Method[Callable[[], Tuple[str]]] = Method(
        RPC.accounts,
        mungers=None,
    )

    @property
    def accounts(self) -> Tuple[str]:
        return self.get_accounts()

    def epoch_munger(
            self,
            epoch_identifier: Optional[EpochIdentifier] = None
    ) -> Tuple[EpochIdentifier]:
        if epoch_identifier is None:
            epoch_identifier = LATEST_MINED
        return (epoch_identifier,)

    epochNumber: Method[Callable[[Optional[EpochIdentifier]], EpochNumber]] = Method(
        RPC.cfx_epochNumber,
        mungers=[epoch_munger],
        result_formatters = get_result_formatters,
    )

    def account_id_munger(
        self,
        account: str,
        epoch_identifier: Optional[EpochIdentifier] = None
    ) -> Tuple[str, EpochIdentifier]:
        if epoch_identifier is None:
            epoch_identifier = self.defaultEpoch
        return (account, epoch_identifier)

    getBalance: Method[Callable[..., Drip]] = Method(
        RPC.cfx_getBalance,
        mungers=[account_id_munger],
        request_formatters= get_request_formatters,
        result_formatters = get_result_formatters,
    )

    getStakingBalance: Method[Callable[..., Drip]] = Method(
        RPC.cfx_getStakingBalance,
        mungers=[account_id_munger],
        request_formatters=get_request_formatters,
        result_formatters=get_result_formatters,
    )

    getCollateralForStorage: Method[Callable[..., int]] = Method(
        RPC.cfx_getCollateralForStorage,
        mungers=[account_id_munger],
        request_formatters=get_request_formatters,
        result_formatters=get_result_formatters,
    )

    getAdmin: Method[Callable[..., str]] = Method(
        RPC.cfx_getAdmin,
        mungers=[account_id_munger],
        request_formatters=get_request_formatters,
    )

    getCode: Method[Callable[..., str]] = Method(
        RPC.cfx_getCode,
        mungers=[account_id_munger],
        request_formatters=get_request_formatters,
    )

    getSponsorInfo: Method[Callable[..., SponsorInfo]] = Method(
        RPC.cfx_getSponsorInfo,
        mungers=[account_id_munger],
        request_formatters=get_request_formatters,
    )

    getNextNonce: Method[Callable[..., int]] = Method(
        RPC.cfx_getNextNonce,
        mungers=[account_id_munger],
        request_formatters=get_request_formatters,
        result_formatters=get_result_formatters,
    )

    getAccount: Method[Callable[..., Account]] = Method(
        RPC.cfx_getAccount,
        mungers=[account_id_munger],
        request_formatters=get_request_formatters,
        result_formatters=get_result_formatters,
    )

    getStatus: Method[Callable[..., ChainStatus]] = Method(
        RPC.cfx_getStatus,
        result_formatters=get_result_formatters,
    )

    def get_block_munger(
        self, block_identifier: BlockIdentifier, full_transactions: bool = False
    ) -> Tuple[BlockIdentifier, bool]:
        return (block_identifier, full_transactions)

    """
    `cfx_getBlockByHash`
    """
    getBlockByHash: Method[Callable[..., BlockData]] = Method(
        RPC.cfx_getBlockByHash,
        mungers=[get_block_munger],
        result_formatters=get_result_formatters,
    )

    """
    `cfx_getBlockByEpochNumber`
    """
    getBlockByEpochNumber: Method[Callable[..., BlockData]] = Method(
        RPC.cfx_getBlockByEpochNumber,
        mungers=[get_block_munger],
        request_formatters=get_request_formatters,
        result_formatters=get_result_formatters,
    )

    getBlocksByEpoch: Method[Callable[..., List[str]]] = Method(
        RPC.cfx_getBlocksByEpoch,
        mungers=[default_root_munger],
        request_formatters=get_request_formatters,
    )

    getBlockRewardInfo: Method[Callable[..., List[RewardInfo]]] = Method(
        RPC.cfx_getBlockRewardInfo,
        mungers=[default_root_munger],
        request_formatters=get_request_formatters,
    )

    getTransactionByHash: Method[Callable[[_Hash32], TxData]] = Method(
        RPC.cfx_getTransactionByHash,
        mungers=[default_root_munger],
        request_formatters=get_request_formatters,
        result_formatters=get_result_formatters,
    )

    # def waitForTransactionReceipt(
    #     self, transaction_hash: _Hash32, timeout: int = 120, poll_latency: float = 0.1
    # ) -> TxReceipt:
    #     try:
    #         return wait_for_transaction_receipt(self.web3, transaction_hash, timeout, poll_latency)
    #     except Timeout:
    #         raise TimeExhausted(
    #             "Transaction {} is not in the chain, after {} seconds".format(
    #                 to_hex(transaction_hash),
    #                 timeout,
    #             )
    #         )

    getTransactionReceipt: Method[Callable[[_Hash32], TxReceipt]] = Method(
        RPC.cfx_getTransactionReceipt,
        mungers=[default_root_munger],
        result_formatters=get_result_formatters,
    )

    def send_transaction_munger(self, transaction: TxParams) -> Tuple[TxParams]:
        # if 'from' not in transaction and is_checksum_address(self.defaultAccount):
        #     transaction = assoc(transaction, 'from', self.defaultAccount)
        #
        # # TODO: move gas estimation in middleware
        # if 'gas' not in transaction:
        #     transaction = assoc(
        #         transaction,
        #         'gas',
        #         get_buffered_gas_estimate(self.web3, transaction),
        #     )
        return (transaction,)

    sendTransaction: Method[Callable[[TxParams], HexBytes]] = Method(
        RPC.cfx_sendTransaction,
        mungers=[send_transaction_munger],
        result_formatters=get_result_formatters,
    )

    sendRawTransaction: Method[Callable[[Union[HexStr, bytes]], HexBytes]] = Method(
        RPC.cfx_sendRawTransaction,
        mungers=[default_root_munger],
        result_formatters=get_result_formatters,
    )

    def call_munger(
        self,
        transaction: TxParams,
        block_identifier: Optional[BlockIdentifier] = None
    ) -> Tuple[TxParams, BlockIdentifier]:
        # if 'from' not in transaction and is_checksum_address(self.defaultAccount):
        #     transaction = assoc(transaction, 'from', self.defaultAccount)

        if block_identifier is None:
            block_identifier = self.defaultEpoch

        return (transaction, block_identifier)

    call: Method[Callable[..., Union[bytes, bytearray]]] = Method(
        RPC.cfx_call,
        mungers=[call_munger]
    )

    def estimate_gas_munger(
        self,
        transaction: TxParams,
        block_identifier: Optional[BlockIdentifier] = None
    ) -> Sequence[Union[TxParams, BlockIdentifier]]:
        # if 'from' not in transaction and is_checksum_address(self.defaultAccount):
        #     transaction = assoc(transaction, 'from', self.defaultAccount)

        if block_identifier is None:
            params: Sequence[Union[TxParams, BlockIdentifier]] = [transaction]
        else:
            params = [transaction, block_identifier]

        return params

    estimateGasAndCollateral: Method[Callable[..., EstimateResult]] = Method(
        RPC.cfx_estimateGasAndCollateral,
        mungers=[estimate_gas_munger],
        request_formatters=get_request_formatters,
        result_formatters=get_result_formatters,
    )

    getLogs: Method[Callable[[FilterParams], List[Log]]] = Method(
        RPC.cfx_getLogs,
        mungers=[default_root_munger],
        request_formatters=get_request_formatters,
        result_formatters=get_result_formatters,
    )

    getBestBlockHash: Method[Callable[..., str]] = Method(
        RPC.cfx_getBestBlockHash,
        mungers=[default_root_munger],
    )

    def populate_transaction(self, tx):
        if "nonce" not in tx:
            tx["nonce"] = self.getNextNonce(tx["from"])
        if "chainId" not in tx:
            status = self.getStatus()
            tx["chainId"] = status["chainId"]
        if "epochHeight" not in tx:
            tx["epochHeight"] = self.epochNumber()
        if "gasPrice" not in tx:
            tx["gasPrice"] = 1
        if "gas" not in tx or "storageLimit" not in tx:
            estimate = self.estimateGasAndCollateral(tx)
            if "gas" not in tx:
                tx["gas"] = estimate["gasLimit"]
            if "storageLimit" not in tx:
                tx["storageLimit"] = estimate["storageCollateralized"]

    # getConfirmationRiskByHash: Method[Callable[..., str]] = Method(
    #     RPC.cfx_getConfirmationRiskByHash,
    #     mungers=[default_root_munger],
    #     result_formatters=get_result_formatters,
    # )